'''
Created on Jun 20, 2017

@author: arnon
'''

from sqlalchemy.engine.url import URL
from sqlalchemy import create_engine
from sqlalchemy.schema import MetaData
from sqlalchemy.orm import sessionmaker, scoped_session
import logging
from .conf_handler import getrootconf
import sys
import sqlite3

mlogger = logging.getLogger(__file__)

URL_ARGS = ['drivername', 'username', 'password', 'host', 'port', 'database', 'query', 'file']
CONNECT_ARGS = ['dialect', 'schema'] + URL_ARGS


def getdbinfo(conf=None, database=None, root='DATABASES', quiet=False, expand=True, expand_map=None):
    databases = getrootconf(conf=conf, root=root, expand=expand, expand_map=expand_map)
    db = None
    if database is not None:
        try:
            db = databases[database]
        except Exception:
            if not quiet:
                raise Exception("Database configuration not found for: %s" % (database))
            else:
                return None
    else:
        db = databases
    return db


def get_item(conf, tag, default=None, ):
    try:
        value = conf[tag]
    except Exception as e:
        if not default:
            raise Exception("missing %s" % (tag)) from e
        value = default
    return value


def sqlite_url(dbconf, database):
    '''
    engine = create_engine('sqlite:///foo.db')
    engine = create_engine('sqlite:////absolute/path/to/foo.db')
    engine = create_engine('sqlite:///C:\\path\\to\\foo.db')
    engine = create_engine(r'sqlite:///C:\path\to\foo.db')
    engine = create_engine('sqlite://')
    '''
    creator = None

    file = dbconf.get('database')
    url_args = get_url_kwargs(dbconf, )
    url = URL(drivername='sqlite', **url_args)

    if not file:
        PY2 = sys.version_info.major == 2
        if PY2:
            params = {}
        else:
            params = {'uri': True}

        url_args_list = ["%s=%s" % (k, v) for k, v in url_args.items()]
        url_args_str = ''
        if url_args_list:
            url_args_str="?%s" % '&'.join(url_args_list)
        creator_uri = 'file::memory:%s' % (url_args_str)
        creator = lambda: sqlite3.connect(creator_uri, **params)

    kwargs = get_rest_kwargs(dbconf, ignore=CONNECT_ARGS, )
    if creator:
        kwargs['creator'] = creator

    return url, kwargs


dialect_map = {
    'sqlite': sqlite_url,
    }


def get_rest_kwargs(dbconf, ignore=[], ):
    kwargs = dict([(k, v) for k, v in dbconf.items()
                   if k not in ignore])
    return kwargs


def get_url_kwargs(dbconf):
    kwargs = dict([(k, v) for k, v in dbconf.items()
                  if k in URL_ARGS and k != 'drivername'])
    return kwargs


class SQLAlchemyConf(object):
    def __init__(self, conf=None, root='DATABASES', database=None, expand=True, expand_map=None, quiet=False, echo=False, logger=None):
        global mlogger

        if logger:
            mlogger = logger
        self.conf = conf
        self.database = database
        self.dbconf = getdbinfo(conf=self.conf, root=root, database=self.database, expand=expand,
                                expand_map=expand_map, quiet=quiet)
        self.engine = None
        self.metadata = None
        self.session = None
        self.echo = echo

    def __repr__(self):
        return repr(self.dbconf)

    def get_info(self):
        return self.dbconf

    def get_url(self):
        dbconf = self.dbconf
        try:
            dialect = get_item(dbconf, 'dialect', )
        except Exception as e:
            raise Exception("dialect key missing in configuration for: %s." % (self.database)) from e

        func = dialect_map.get(dialect)

        if func:
            mlogger.debug('SQLAlchemyConf: specialized dialect map for {}: {}.'
                          .format(dialect, getattr(func, "__name__", repr(func))))
            url, kwargs = func(dbconf, self.database)
        else:
            mlogger.debug('SQLAlchemyConf: default SQLAlchemy mapping for {}.'.format(dialect))
            url_args = get_url_kwargs(dbconf, )
            try:
                url = URL(drivername=dialect, **url_args)
                # fetch other connection args (called query in sqlalchemy.engine.url.URL
            except Exception as e:
                raise Exception('Failed to create URL for : {}:{}.'
                                .format(dialect, url_args.get('drivername'))) from e
            kwargs = get_rest_kwargs(dbconf, ignore=CONNECT_ARGS, )
        mlogger.debug("SQLAlchemyConf: url: {}, kwargs: {}.".format(url, kwargs))
        return url, kwargs

    def get_engine(self, force=False,):
        if self.engine and not force:
            mlogger.debug("SQLAlchemyConf: engine already set and force is not set.")
            return self.engine
        mlogger.debug("SQLAlchemyConf: getting SQLAlchemy URL.")
        url, kwargs = self.get_url()
        self.engine = engine = create_engine(url, **kwargs, echo=self.echo)
        return engine

    def get_metadata(self, force=False):
        if self.metadata and not force:
            return self.metadata
        dbconf = self.dbconf
        engine = self.get_engine(force=force)
        schema = dbconf.get('schema')
        self.metadata = metadata = MetaData(bind=engine, schema=schema)
        return metadata

    def get_session(self, force=False, autocommit=True):
        if self.session and not force:
            return self.session
        session_factory = sessionmaker(bind=self.get_metadata(force=force).bind.engine,
                                       autocommit=autocommit)
        Session = scoped_session(session_factory)
        self.session = session = Session()
        # session=Session()
        return session


if __name__ == '__main__':
    dbconf = {
        "EVENTOR": {
            "DATABASES": {
                "default": {
                    "dialect": "sqlite",
                    "query": {
                        "cache": "shared"
                    },
                },
                "sqfile00": {
                    "dialect": "sqlite",
                    "database": "/var/acrisel/sand/eventor/eventor/eventor/examples/example00.db",
                },
                "pgdb1": {
                    "dialect": "postgresql",
                    "drivername": "psycopg2",
                    "username": "arnon",
                    "password": "arnon42",
                    "host": "localhost",
                    "port": 5433,
                    "database": "pyground",
                    "schema": "play",
                },
                "pgdb2": {"dialect": "postgresql",
                          "drivername": "psycopg2",
                          "username": "arnon",
                          "password": "arnon42",
                          "host": "192.168.1.70",
                          "port": 5432,
                          "database": "pyground",
                          "schema": "play",
                          }}}}

    for db in ['default', 'sqfile00', 'pgdb1', 'pgdb2', ]:
        sqlconf = SQLAlchemyConf(dbconf, root='EVENTOR.DATABASES', database=db)
        url, connect_args = sqlconf.get_url()
        print(db, url, connect_args)

    from .dbschema import info_table
    from sqlalchemy.ext.declarative import declarative_base
    sqlconf = SQLAlchemyConf(dbconf, root='EVENTOR.DATABASES', database='default', echo=True)
    metadata = sqlconf.get_metadata()
    engine = sqlconf.get_engine()
    Base = declarative_base(bind=engine, metadata=metadata)
    Info = info_table(Base)

    session = sqlconf.get_session(force=True, autocommit=False)
    # info_table = metadata.tables['Info']
    # print("Schema:", info_table.schema)

    # print("metadata tables:")
    # for name, table in metadata.tables.items():
    #        print("    {}: {}".format(name, table))

    metadata.create_all()
    # session.flush()
    # session.commit()

    data = {
        'app_version': '5.1.0',
        }
    for name, value in data.items():
        db_info = Info(run_id='test1', name=name, value=value)
        session.add(db_info)
    session.commit()



#!/usr/bin/env python3

# adopted from https://gist.github.com/bortzmeyer/1284249

# Here is the right solution today:

import subprocess


def sshcmd(host, command, user=None, check=False):
    ''' Runs ssh command via subprocess.  Assuming .ssh/config is configured.

    Args:
        host: target host to send command to
        command: command to run on host
        user: (optional) user to use to login to host
        check: pass to subprocess.run; if set, checks return code and raises subprocess.CalledProcessError if none-zero result

    Returns:
        subprocess.CompletedProcess object
    '''

    where = "%s" % host if user is None else "%s@%s" %(user, host)
    result = subprocess.run(["ssh", where, command],
                            shell=False,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=check)
    return result


if __name__ == '__main__':
    out=sshcmd("ubly", "ls -l", check=False).stdout.decode()
    print(out)



#!/usr/bin/env python
'''
Created on Aug 30, 2017

@author: arnon
'''

from eventor.lib.engine import Eventor
import pickle
import sys
import struct
import importlib.util
import multiprocessing as mp
from acrilog import SSHLogger as Logger
from acrilog import SSHLoggerClientHandler
import sshpipe as sp
import logging
import pprint
import yaml
from copy import copy
import os
import select

mlogger = None


class EventorAgentError(Exception):
    pass


level_formats = {logging.DEBUG: ("[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ]"
                                 "[ %(processName)-11s ][ %(message)s ]"
                                 "[ %(module)s.%(funcName)s(%(lineno)d) ]"),
                 'default': ("[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ]"
                             "[ %(processName)-11s ][ %(message)s ]"),
                 }

RECOVER_ARGS_DIR = '/tmp'
RECOVER_ARGS_FILE = 'eventor_agent_args'


def new_recvoer_args_file(name=None):
    file = "{}.".format(name) if name else ""
    file = "{}{}.{}.dat".format(file, RECOVER_ARGS_FILE, os.getpid())
    file = os.path.join(RECOVER_ARGS_DIR, file)
    return file


def last_recvoer_args_file():
    files = filter(lambda x: os.path.isfile(x) and x.startswith(RECOVER_ARGS_FILE), os.listdir(RECOVER_ARGS_DIR))
    files = [os.path.join(RECOVER_ARGS_DIR, f) for f in files]  # add path to each file
    files.sort(key=lambda x: os.path.getmtime(x))
    return files[-1] if len(files) > 0 else None


def cmdargs(args=None):
    import argparse

    filename = os.path.basename(__file__)
    progname = filename.rpartition('.')[0]

    parser = argparse.ArgumentParser(prog=progname, description="runs EventorAgent object.")
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',
                                       help='additional help')
    parser_act = subparsers.add_parser('action', aliases=['act'], help='perform new action')
    parser_rec = subparsers.add_parser('recover', aliases=['rec'], help='recover previous act')

    parser_act.set_defaults(run=True, recover=False)
    parser_rec.set_defaults(run=False, recover=True)

    parser_act.add_argument('--imports', type=str, required=False, dest='imports', nargs='*',
                            help="""import module before pickle loads.""")
    parser_act.add_argument('--host', type=str,
                            help="""Host on which this command was sent to.""")
    parser_act.add_argument('--ssh-server-host', type=str, dest='ssh_host',
                            help="""SSH Host to use for back channel.""")
    parser_act.add_argument('--log-info', type=str, dest='log_info',
                            help="""Logger info dictionary json coded.""")
    parser_act.add_argument('--file', type=str, required=False,
                            help=("File to store or recover memory. With --pipe, it would store"
                                  " memory into file. Without --pipe, it would recover memory from"
                                  " store"))
    parser_act.add_argument('--pipe', action='store_true',
                            help=("Indicates that memory should be read from STDIN. If --pipe not"
                                  " provided, --file must be."))
    parser_act.add_argument('--debug', action='store_true',
                            help="Invokes additional debug utilities, e.g., store args for recovery.")

    parser_rec.add_argument('--file', type=str, required=False,
                            help="""File from which to restore previous args""")
    pargs = parser.parse_args(args=args)

    assert pargs.file is not None or pargs.pipe, "--pipe or --file must be provided."

    return pargs


def do_imports(imports):
    imports = imports_from_cmd(imports)
    mlogger.debug("Importing {}.".format(imports))
    for import_file, import_modules in imports:
        if not import_file:
            for module in import_modules:
                mlogger.debug("Importing %s." % (module))
                try:
                    from importlib import import_module
                    import_module(module)
                except Exception as e:
                    mlogger.critical("Failed to import: %s." % (module))
                    mlogger.exception(e)
                    # signal to parent via stdout
                    print('TERM')
                    print(e, file=sys.stderr)
                    return
        else:
            for module in import_modules:
                mlogger.debug("Importing %s from %s." % (module, import_file))
                try:
                    spec = importlib.util.spec_from_file_location(module, import_file)
                    spec_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(spec_module)
                    # sys.modules[import_module] = module # not needed for now
                except Exception as e:
                    mlogger.critical("Failed to import: %s %s;" % (import_module, import_file))
                    mlogger.exception(e)
                    # signal to parant via stdout
                    print('TERM')
                    print(e, file=sys.stderr)
                    return


def start_eventor(queue, logger_info, **kwargs):
    global mlogger
    mlogger = Logger.get_logger(logger_info,)  # logger_info['name'])
    mlogger.debug('Starting EventorAgent:\n{}'.format(pprint.pformat(kwargs, indent=4)))
    try:
        eventor = Eventor(**kwargs)
    except Exception as e:
        mlogger.critical("Failed to start EventorAgent.")
        mlogger.exception(e)
        queue.put(('TERM', e))
        return

    mlogger.debug('Initiated EventorAgent object, going to run().')

    try:
        eventor.run()
    except Exception as e:
        mlogger.critical("Failed to run EventorAgent, passing TERM to main process.")
        mlogger.exception(e)
        queue.put(('TERM', e))
    else:
        mlogger.debug('EventorAgent finished: passing DONE to main process.')
        queue.put(('DONE', ''))


def pull_from_pipe(queue=None,):
    ''' Pipe listener wait for one message.  Once receive (DONE or TERM) it ends.
    '''
    global mlogger

    def set_result(msg, e):
        if queue is not None:
            queue.put((msg, e))
            return
        else:
            return (msg, e)

    # in this case, whiting for possible termination message from server
    try:
        msgsize_raw = sys.stdin.buffer.read(4)
    except Exception as e:
        mlogger.critical('Failed to read STDIN.')
        mlogger.exception(e)
        return set_result('TERM', e)
    try:
        msgsize = struct.unpack(">L", msgsize_raw)
    except Exception as e:
        mlogger.critical('Failed pickle loads message size from STDIN; received: %s' % hex(msgsize_raw))
        mlogger.exception(e)
        return set_result('TERM', e)
    try:
        msg_pack = sys.stdin.buffer.read(msgsize[0])
    except Exception as e:
        mlogger.critical('Failed to read STDIN.')
        mlogger.exception(e)
        return set_result('TERM', e)
    try:
        msg = pickle.loads(msg_pack)
    except Exception as e:
        mlogger.critical('Failed pickle loads message from STDIN.')
        mlogger.exception(e)
        return set_result('TERM', e)

    mlogger.debug('Received message from remote parent: {}; passing to main process.'.format(msg))

    return set_result(msg, '')


def imports_from_cmd(imports_str):
    imports = dict()

    for import_str in imports_str:
        import_file, _, import_modules_str = import_str.partition(':')
        import_modules = import_modules_str.split(':')
        file_modules = imports.get(import_file, list())
        file_modules.extend(import_modules)
        imports[import_file] = file_modules
    imports = [(import_file, set(modules)) for import_file, modules in imports.items() if modules]
    return imports


def check_agent_process(agent,):
    if not agent.is_alive():
        agent.join()
        mlogger.debug("Agent is not alive! terminating.")
        print('TERM')
        print("Agent aborted unexpectedly.", file=sys.stderr)
        return False
    return True


def logger_remote_handler(logger_queue, log_info_recv, ssh_host,):
    try:
        remote_logger_handler = SSHLoggerClientHandler(logger_info=log_info_recv, ssh_host=ssh_host)
    except Exception as e:
        raise EventorAgentError("Failed to create SSHLoggerClientHandler on: {}; {}"
                                .format(ssh_host), repr(e))

    listener = logging.handlers.QueueListener(logger_queue, remote_logger_handler)
    listener.start()
    return listener


class EventorAgentQueueHandler(logging.handlers.QueueHandler):
    def __init__(self, queue):
        super(EventorAgentQueueHandler, self).__init__(queue)
        self.emitter = open("/var/log/eventor/eventor_agent_queue_handler_records.log", 'w')
        self.count = 0

    def emit(self, record):
        self.count += 1
        self.emitter.write("{}: \n".format(repr(record.__dict__)))
        logging.handlers.QueueHandler.emit(self, record)

    def __del__(self):
        self.emitter.close()


class SSHPipeEventorAgent(sp.SSHPipeHandler):
    ''' SSHPipeSocketHandler modeled over logging.handlers.SocketHandler
    '''

    def __init__(self, agent_args, *args, **kwargs):
        super(sp.SSHPipeHandler, self).__init__(*args, **kwargs)
        self.agent_args = agent_args
        self.mlogger.debug("EventorAgent params:\n{}.".format(self.args))

    def atstart(self, receieved):
        args = self.agent_args
        log_info, imports, host, ssh_host, pipe, file = \
        args.log_info, args.imports, args.host, args.ssh_host, args.pipe, args.file

    def atexit(self, received):
        # if self.file is not None:
        #     self.file.close()
        super(sp.SSHPipeHandler, self).atexit(received)
        self.mlogger.debug("Closed handlers.")

    def handle(self, received):
        """
        Send a pickled string to the socket.

        This function allows for partial sends which can happen when the
        network is busy.
        """
        # self.file.write(str(received))
        self.emit(received)
        self.mlogger.debug("Emitted record to socket: {}.".format(received))


def run(args):
    client = SSHPipeEventorAgent(args)
    client.service_loop()


def initiate_logger(log_info):
    global mlogger
    log_info_recv = yaml.load(log_info)
    handler_kwargs = log_info_recv['handler_kwargs']
    logger_info_local = copy(log_info_recv)
    del logger_info_local['port']
    del logger_info_local['handler_kwargs']

    logger_kwargs = {}
    logger_kwargs.update(logger_info_local)
    logger_kwargs.update(handler_kwargs)
    logger_queue = mp.Queue()
    queue_handler = logging.handlers.QueueHandler(logger_queue)
    logger = Logger(console=False, handlers=[queue_handler], **logger_kwargs)
    logger.start()

    logger_info = logger.logger_info()
    mlogger = Logger.get_logger(logger_info=logger_info)

    return logger, logger_info_local, logger_queue, log_info_recv, logger_info


def run_eventor(args):
    ''' Runs EventorAgent in remote host.

    Args:
        log_info
        imports
        host
        file
        pipe
    '''
    global mlogger

    run, recovery = args.run, args.recover

    if run:
        log_info, imports, host, ssh_host, pipe, file, debug = \
            args.log_info, args.imports, args.host, args.ssh_host, args.pipe, args.file, args.debug
    else:  # recovery:
        # this is recovery: read args from file
        file = args.file
        with open(file, 'rb') as f:
            args = pickle.load(f)
            memory = pickle.load(f)
            # Dont take pipe, and file from arg store
            log_info, imports, host, ssh_host, debug = \
                args.log_info, args.imports, args.host, args.ssh_host, args.debug

    logger, logger_info_local, logger_queue, log_info_recv, logger_info = initiate_logger(log_info)

    mlogger.debug("Starting agent: {}.".format(args))

    # In case of debug, store arguments.
    # but, also, it is not recovery (pipe is provided)
    if debug and not recovery and file:
        file = new_recvoer_args_file(file)
        try:
            with open(file, 'wb') as f:
                pickle.dump(args, f)
        except Exception as e:
            mlogger.exception(e)
            mlogger.debug("Failed to store args.")
            close_run(logger=logger, msg="TERM", err=e)
            return

        mlogger.debug("Stored agent args: {}".format(file))

    mlogger.debug('Local logger:\n{}'.format(logger_info_local))
    mlogger.debug('Module logger:\n{}'.format(log_info))

    try:
        dipatcher_to_remote_logger = logger_remote_handler(logger_queue, log_info_recv=log_info_recv, ssh_host=ssh_host,)  # logdir=handler_kwargs['logdir'])
    except Exception as e:
        mlogger.exception(e)
        mlogger.debug("Failed to to create remote logger on: {}.".format(ssh_host))
        close_run(logger=logger, msg="TERM", err=e)
        return

    mlogger.debug('Started logger_remote_handler: {}.'.format(ssh_host))

    if imports is not None:
        do_imports(imports)

    if run:
        mlogger.debug("Fetching workload. from pipe.")

        memory, e = pull_from_pipe()
        if e:
            mlogger.critical("Failed to get Eventor memory from pipe.")
            close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=e)
            return

        # store memory into file
        if not recovery and file and debug:
            mlogger.debug("Storing workload to {}.".format(file))
            try:
                with open(file, 'ab') as file:
                    pickle.dump(memory, file)
            except Exception as e:
                mlogger.critical("Failed to pickle dump workload to {}.".format(file))
                mlogger.exception(e)
                close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=e)
                return
    '''
    else:
        mlogger.debug("Fetching workload from file.")
        try:
            with open(file, 'rb') as file:
                memory = pickle.load(file)
        except Exception as e:
            mlogger.critical("Failed to pickle load workload from {}.".format(file))
            mlogger.exception(e)
            close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=e)
            return
    '''
            
    mlogger.debug("Memory received:\n{}".format(pprint.pformat(memory, indent=4, )))

    try:
        kwargs = memory.kwargs.copy()
        # change logger_info to this agent logger info.
        memory.logger_info = logger_info
        kwargs['host'] = host
        kwargs['memory'] = memory
    except Exception as e:
        mlogger.critical("Failed get kwargs from received memory.")
        mlogger.exception(e)
        close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=e)
        return

    mlogger.debug("Starting Eventor subprocess on remote host.") #:\n%s" % pprint.pformat(kwargs, indent=4))

    child_q = mp.Queue()
    eventor_listener_q = mp.Queue()
    kwargs['listener_q'] = eventor_listener_q

    try:
        agent = mp.Process(target=start_eventor, args=(child_q, logger_info,), kwargs=kwargs, daemon=False)
        agent.start()
    except Exception as e:
        mlogger.critical("Failed to start Eventor process.")
        mlogger.exception(e)
        close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=e)
        return

    mlogger.debug("Eventor subprocess pid: {}".format(agent.pid))

    # wait for remote parent or from child Eventor
    if not check_agent_process(agent):
        mlogger.debug("Agent process is dead, exiting. pid: {}".format(agent.pid))
        close_run(dipatcher_to_remote_logger, logger, )  # msg='TERM', err=error)
        return

    mlogger.debug("Starting EventorAgent select loop.")
    loop_is_active = True
    while loop_is_active:
        rselected, _, _ = select.select([child_q._reader, sys.stdin],[],[])
        mlogger.debug("Returned from select: {}.".format(repr(rselected)))

        for ioitem in rselected:
            if ioitem == sys.stdin:
                result = pull_from_pipe()
                mlogger.debug("Received from select remote parent: {}.".format(result))
                if result:
                    # received FINISH, STOP, or QUIT message from parent
                    loop_is_active = False
            elif ioitem == child_q._reader:
                result = child_q.get(timeout=0.5)
                mlogger.debug("Received from select local Eventor: {}.".format(result))
                if result:
                    agent.join()
                    loop_is_active = False

    msg, error = result
    mlogger.debug("Pulled message from control queue: message: {}; error: {}.".format(msg, error,))
    if msg == 'DONE':
        # msg from child - eventor agent is done
        mlogger.debug("Joining with Eventor process after: {}.".format(msg))
        agent.join()
        close_run(dipatcher_to_remote_logger, logger, )
    elif msg == 'TERM':
        # TODO: need to change message from parent to STOP - not TERM
        # got message to quit, need to kill primo process and be done
        # Well since process is daemon, it will be killed when parent is done
        eventor_listener_q.put('STOP')
        mlogger.debug("Joining with Eventor process after: {}.".format(msg))
        agent.join()
        close_run(dipatcher_to_remote_logger, logger, msg='TERM', err=error)
        # TODO(Arnon): how to terminate listener that is listening
    elif msg in ['STOP', 'FINISH']:
        # TODO: need to change message from parent to STOP - not TERM
        # got message to quit, need to kill primo process and be done
        # Well since process is daemon, it will be killed when parent is done
        eventor_listener_q.put('STOP')
        mlogger.debug("Joining with Eventor process after: {}.".format(msg))
        agent.join()
        result = child_q.get(timeout=0.5)
        mlogger.debug("Eventor process joint result: {}.".format(result))
        close_run(dipatcher_to_remote_logger, logger, msg='DONE', err=error)


def close_run(listener=None, logger=None, msg=None, err=None):
    global mlogger

    if msg is not None:
        print(msg)
    if err is not None:
        print(err, file=sys.stderr)

    mlogger.debug("Closing EventorAgent dipatcher_to_remote_logger.")
    if listener:
        listener.stop()

    mlogger.debug("Closing EventorAgent logger.")
    if logger:
        logger.stop()


'''
def recover_eventor(args):
    file = args.file
    if file is None:
        file = last_recvoer_args_file()
    if file is not None:
        with open(file, 'rb') as f:
            saved_args = pickle.load(f)
        run_eventor(saved_args)
    else:
        raise EventorAgentError("Trying to run in recovery, but no recovery file found.")
'''

if __name__ == '__main__':
    # mp.freeze_support()
    # mp.set_start_method('spawn')
    
    
    args = cmdargs()
    '''
    if args.run:
        run_eventor(args)
    else:
        recover_eventor(args)
    '''
    run_eventor(args)
    '''
    args = cmdargs(['rec', '--file', 'myfile'])
    print(args)
    '''
    



__version__ = '5.1.8'
__db_version__ = '1.0.0'

from eventor.lib.engine import Eventor, get_unique_run_id
from eventor.lib.eventor_types import AssocType, TaskStatus, DbMode, StepStatus, StepReplay
from eventor.lib.eventor_types import RunMode, Invoke
from eventor.lib.event import or_
from eventor.lib.utils import calling_module, store_from_module
from eventor.lib.conf_handler import merge_configs

STEP_RERUN = StepReplay.rerun  # reruns step regardless if previously succeeded
STEP_SKIP = StepReplay.skip  # skip step if previously succeeded

STEP_READY = StepStatus.ready  # step is ready
STEP_ACTIVE = StepStatus.active  # step is running
STEP_SUCCESS = StepStatus.success  # step succeeded
STEP_FAILURE = StepStatus.failure  # step failed
STEP_COMPLETE = StepStatus.complete  # step complete with success or failure

RUN_RESTART = RunMode.restart  # run flow from start
RUN_RECOVER = RunMode.recover  # reruns failed steps

# Note: internal use only
RUN_CONTINUE = RunMode.continue_  # continue from where it left in previous loop



'''
Created on Oct 31, 2016

@author: arnon
'''

from .step import Step
import logging
from .event import Event
from .eventor_types import EventorError

mlogger = logging.getLogger(__name__)


class Assoc(object):

    def __init__(self, event, assoc_obj):
        if not isinstance(event, Event):
            raise EventorError('event argument must be Event, but found %s' % type(event))
        if not isinstance(assoc_obj, (Event, Step)):
            raise EventorError('assoc_obj argument must be Event or Step, but found %s' % type(assoc_obj))

        self.event = event
        self.assoc_obj = assoc_obj

    def __repr__(self):
        return "Assoc(%s, %s)" % (repr(self.event), repr(self.assoc_obj))

    def __str__(self):
        return "Assoc(%s, %s)" % (str(self.event), str(self.assoc_obj))




# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import eventor as evr
import logging
import os
from acris import virtual_resource_pool as vrp

appname = os.path.basename(__file__)


class StepResource(vrp.Resource):
    pass


def prog(progname):
    logger = logging.getLogger(os.getenv("EVENTOR_LOGGER_NAME"))
    logger.info("doing what %s is doing" % progname)
    logger.info("EVENTOR_STEP_SEQUENCE: %s" % os.getenv("EVENTOR_STEP_SEQUENCE"))
    return progname


rp1 = vrp.ResourcePool('rp1', resource_cls=StepResource, policy={'resource_limit': 2, }).load()

# ev=evr.Eventor(store=':memory:', logging_level=logging.INFO)
config = os.path.abspath('runly.conf')
ev = evr.Eventor(name=appname, config=config, config_tag='EVENTOR')

ev1s = ev.add_event('run_step1')
ev2s = ev.add_event('run_step2')
ev3s = ev.add_event('run_step3')

s1 = ev.add_step('s1', func=prog, kwargs={'progname': 'prog1'},
                 triggers={evr.STEP_SUCCESS: (ev2s,)},
                 acquires=[(rp1, 1)])
s2 = ev.add_step('s2', func=prog, kwargs={'progname': 'prog2'},
                 triggers={evr.STEP_SUCCESS: (ev3s,)},
                 acquires=[(rp1, 1)])
s3 = ev.add_step('s3', func=prog, kwargs={'progname': 'prog3'},
                 acquires=[(rp1, 1)])

ev.add_assoc(ev1s, s1)
ev.add_assoc(ev2s, s2)
ev.add_assoc(ev3s, s3)

ev.trigger_event(ev1s, 1)
ev.run()
ev.close()



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import eventor as evr
import logging
import os

appname = os.path.basename(__file__)


def prog(progname, success=True):
    logger = logging.getLogger(os.getenv("EVENTOR_LOGGER_NAME"))
    logger.info("{} doing what %s is doing: {}".format(progname, success))
    logger.info("EVENTOR_STEP_SEQUENCE: %s" % os.getenv("EVENTOR_STEP_SEQUENCE"))
    if not success:
        raise RuntimeError("Instigating failure.")
    return progname


def build_flow(run_mode, run_id=None, success=True):
    ev = evr.Eventor(name=appname, run_mode=run_mode, run_id=run_id,
                     config={'LOGGING':
                             {'logging_level': logging.INFO}})

    ev1s = ev.add_event('run_step1')
    ev2s = ev.add_event('run_step2')
    ev3s = ev.add_event('run_step3')

    kwprog1 = {'progname': 'prog1'}
    kwprog2 = {'progname': 'prog2', 'success': success}
    kwprog3 = {'progname': 'prog3'}

    s1 = ev.add_step('s1', func=prog, kwargs=kwprog1, triggers={evr.STEP_SUCCESS: (ev2s, )})
    s2 = ev.add_step('s2', func=prog, kwargs=kwprog2, triggers={evr.STEP_SUCCESS: (ev3s,)})
    s3 = ev.add_step('s3', func=prog, kwargs=kwprog3,)

    ev.add_assoc(ev1s, s1, delay=1)
    ev.add_assoc(ev2s, s2, delay=1)
    ev.add_assoc(ev3s, s3, delay=1)

    ev.trigger_event(ev1s, 1)
    return ev


def construct_and_run(recovery=True):
    # Instigate failure in run
    ev = build_flow(run_mode=evr.RUN_RESTART, success=not recovery,)
    run_id = ev.run_id
    ev.run()
    ev.close()

    if recovery:
        # recover run
        ev = build_flow(run_mode=evr.RUN_RECOVER, run_id=run_id)
        ev.run()
        ev.close()


if __name__ == '__main__':
    construct_and_run(recovery=True)



import os
import sys

from setuptools import setup
from distutils.sysconfig import get_python_lib
import importlib.util

''' 
is the Python package in your project. It's the top-level folder containing the 
__init__.py module that should be in the same directory as your setup.py file
/-
  |- README.rst
  |- CHANGES.txt
  |- setup.py
  |- dogs 
     |- __init__.py
     |- catcher.py

To create package and upload:

  python setup.py register
  python setup.py sdist
  twine upload -s dist/path/to/gz

'''


def import_setup_utils():
    # load setup utils
    try:
        setup_utils_spec = \
            importlib.util.spec_from_file_location("setup.utils",
                                                   "setup_utils.py")
        setup_utils = importlib.util.module_from_spec(setup_utils_spec)
        setup_utils_spec.loader.exec_module(setup_utils)
    except Exception as err:
        raise RuntimeError("Failed to find setup_utils.py."
                           " Please copy or link.") from err
    return setup_utils


setup_utils = import_setup_utils()
HERE = os.path.abspath(os.path.dirname(__file__))
PACKAGE = "eventor"
NAME = PACKAGE
METAPATH = os.path.join(HERE, PACKAGE, "__init__.py")
metahost = setup_utils.metahost(PACKAGE)

'''
DESCRIPTION is just a short description of your project. A sentence will suffice.
'''
DESCRIPTION = '''eventor is a python programming facility to program event based sequence of activities'''

'''
AUTHOR and AUTHOR_EMAIL are what they sound like: your name and email address. This 
information is optional, but it's good practice to supply an email address if people 
want to reach you about the project.
'''
AUTHORS = 'Acrisel Team'
AUTHORS_EMAIL = 'support@acrisel.com'

'''
URL is the URL for the project. This URL may be a project website, the Github repository, 
or whatever URL you want. Again, this information is optional.
'''
URL = 'https://github.com/Acrisel/eventor'

# version_file=os.path.join(PACKAGE, 'VERSION.py')
# with open(version_file, 'r') as vf:
#    vline=vf.read()
# VERSION = vline.strip().partition('=')[2].replace("'", "")
# VERSION =__import__(version_file).__version__
VERSION = setup_utils.read_version(metahost=metahost)

# Warn if we are installing over top of an existing installation. This can
# cause issues where files that were deleted from a more recent Accord are
# still present in site-packages. See #18115.
overlay_warning = False
if "install" in sys.argv:
    existing_path = setup_utils.existing_package(PACKAGE)

scripts = setup_utils.scripts(PACKAGE)

# Find all sub packages
packages = setup_utils.packages(PACKAGE)
required = setup_utils.read_required(metahost=metahost)

setup_info = {
    'name': NAME,
    'version': VERSION,
    'url': URL,
    'author': AUTHORS,
    'author_email': AUTHORS_EMAIL,
    'description': DESCRIPTION,
    'long_description': open("README.rst", "r").read(),
    'license': 'MIT',
    'keywords': 'project, virtualenv, parameters',
    'packages': packages,
    'scripts': scripts,
    'install_requires': [
        'SQLAlchemy>=1.1.5',
        'acrilib>=1.0.2',
        'acris>=2.0.2',
        'acrilog>=2.0.11',
        'pyyaml>=3.12',
        'namedlist>=1.7',
        ],
    'extras_require': {'extra': ['setproctitle>=1.1']},
    'classifiers': ['Development Status :: 5 - Production/Stable',
                    'Environment :: Other Environment',
                    # 'Framework :: Project Settings and Operation',
                    'Intended Audience :: Developers',
                    'License :: OSI Approved :: MIT License',
                    'Operating System :: OS Independent',
                    'Programming Language :: Python',
                    'Programming Language :: Python :: 3',
                    'Programming Language :: Python :: 3.2',
                    'Programming Language :: Python :: 3.3',
                    'Programming Language :: Python :: 3.4',
                    'Programming Language :: Python :: 3.5',
                    'Programming Language :: Python :: 3.6',
                    'Topic :: Software Development :: Libraries :: Application Frameworks',
                    ]}
setup(**setup_info)


if overlay_warning:
    sys.stderr.write("""

========
WARNING!
========

You have just installed ProjEnv over top of an existing
installation, without removing it first. Because of this,
your install may now include extraneous files from a
previous version that have since been removed from
Accord. This is known to cause a variety of problems. You
should manually remove the

%(existing_path)s

directory and re-install ProjEnv.

""" % {"existing_path": existing_path})



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import eventor as evr
import logging
import os
from acris import virtual_resource_pool as rp
from eventor.examples.run_types import prog, Container, StepResource

# logger=logging.getLogger(__name__)

db = 'pgdb2'
config=os.path.abspath('runly.conf')
# because OSX adds /var -> /private/var
if config.startswith('/private'):
    config = config[8:]
ev = evr.Eventor(name=os.path.basename(__file__), config=config, config_tag='EVENTOR', store=db, ) #config={'sleep_between_loops': 1}) # store=':memory:',

ev0first = ev.add_event('s0_start')
ev0next = ev.add_event('s0_next')
ev0end = ev.add_event('s0_end')
ev00first = ev.add_event('s0_00_start')
ev00next = ev.add_event('s0_s00_next')
ev00end = ev.add_event('s0_s00_end')
ev1s = ev.add_event('s0_s00_s1_start')
ev1success = ev.add_event('s0_s00_s1_success')
ev2s = ev.add_event('s0_s00_s2_start', expr=(ev1success,))
ev2success = ev.add_event('s0_s00_s2_success')
ev3s = ev.add_event('s0_s00_s3_start', expr=(ev2success,))

rp1 = rp.ResourcePool('RP1', resource_cls=StepResource, policy={'resource_limit': 2, }).load()                   
rp2 = rp.ResourcePool('RP2', resource_cls=StepResource, policy={'resource_limit': 2, }).load()                   

metaprog1 = Container(progname='s0', loop=[1,2,], max_concurrent=2, iter_triggers=(ev00first,), end_triggers=(ev0end,))
s0first = ev.add_step('s0_start', func=metaprog1, kwargs={'initial': True, }, acquires=[(rp1, 1), ], releases=[], 
                    config={'max_concurrent': -1, 'task_construct': 'invoke', 'pass_logger_to_task': True})
s0next = ev.add_step('s0_next', func=metaprog1, config={'task_construct': 'invoke', 'pass_logger_to_task': True})
s0end = ev.add_step('s0_end', releases=[(rp1, 1), ], config={'task_construct': 'invoke', 'pass_logger_to_task': True})
  
metaprog2 = Container(progname='00', loop=[1,2,], max_concurrent=2, iter_triggers=(ev1s,), end_triggers=(ev00end,))
s00first = ev.add_step('s0_s00_start', func=metaprog2, kwargs={'initial': True, }, acquires=[(rp2, 1), ], releases=[], 
                     config={'max_concurrent': -1, 'task_construct': 'invoke', 'pass_logger_to_task': True})
s00next = ev.add_step('s0_s00_next', func=metaprog2, config={'task_construct': 'invoke', 'pass_logger_to_task': True})
s00end = ev.add_step('s0_s00_end', releases=[(rp2, 1),], config={'task_construct': 'invoke', 'pass_logger_to_task': True}, triggers={evr.StepStatus.success: (ev0next,), })

s1 = ev.add_step('s0.s00.s1', func=prog, kwargs={'progname': 'prog1'}, triggers={evr.StepStatus.success: (ev1success,),}) 
s2 = ev.add_step('s0.s00.s2', func=prog, kwargs={'progname': 'prog2'}, triggers={evr.StepStatus.success: (ev2success,), }, host='ubuntud01_eventor')
s3 = ev.add_step('s0.s00.s3', func=prog, kwargs={'progname': 'prog3'}, triggers={evr.StepStatus.complete: (ev00next,), })

ev.add_assoc(ev0first, s0first)
ev.add_assoc(ev0next, s0next)
ev.add_assoc(ev0end, s0end)
ev.add_assoc(ev00first, s00first)
ev.add_assoc(ev00next, s00next)
ev.add_assoc(ev00end, s00end)
ev.add_assoc(ev1s, s1)
ev.add_assoc(ev2s, s2)
#ev.add_assoc(ev00end, s0next)
ev.add_assoc(ev3s, s3)

ev.trigger_event(ev0first)
ev.run()
ev.close()


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
"""

About
=========
:synopsis:     example use of grapior
:moduleauthor: Arnon Sela
:date:         Oct 18, 2016
:description:  use gradior dependencies and recovery

Outputs:
-------------------
N/A

Dependencies:
-------------------
N/A

**History:**
-------------------

:Author: Arnon Sela
:Modification:
   - Initial entry
:Date: Oct 18, 2016


API DOC:
===============
"""

import eventor as evr
import logging
import os
from eventor.examples.run_types import prog

logger = logging.getLogger(__file__)


def construct_and_run():
    # db = 'sqfile00'
    db = 'pgdb2'
    config = os.path.abspath('runly.conf')
    # because OSX adds /var -> /private/var
    if config.startswith('/private'):
        config = config[8:]
    # TODO: assume import_module is __file__ if not provided
    ev = evr.Eventor(name=os.path.basename(__file__), config=config, config_tag='EVENTOR', store=db)

    ev1s = ev.add_event('run_step1')
    ev2s = ev.add_event('run_step2')
    ev3s = ev.add_event('run_step3')

    host = 'ubuntud01_eventor'
    # host = 'ubuntud01'
    s1 = ev.add_step('s1', func=prog, kwargs={'progname': 'prog1'},
                     triggers={evr.STEP_SUCCESS: (ev2s,)})
    s2 = ev.add_step('s2', func=prog, kwargs={'progname': 'prog2'}, host=host,
                     triggers={evr.STEP_SUCCESS: (ev3s,)})
    s3 = ev.add_step('s3', func=prog, kwargs={'progname': 'prog3'})

    ev.add_assoc(ev1s, s1)
    ev.add_assoc(ev2s, s2)
    ev.add_assoc(ev3s, s3)

    ev.trigger_event(ev1s, '1')
    ev.run()
    # ev.close()


if __name__ == '__main__':
    # import multiprocessing as mp
    # mp.freeze_support()
    # mp.set_start_method('spawn')
    construct_and_run()



'''
Created on Oct 21, 2016

@author: arnon
'''

from sqlalchemy import Column, UniqueConstraint, Sequence, create_engine
from sqlalchemy import Integer, BigInteger, String, PickleType, DateTime, Boolean
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime
from sqlalchemy import Enum as SQLEnum

from .eventor_types import TaskStatus
from namedlist import namedlist


Info = namedlist("InfoRow", "id_ run_id name value")


def info_from_db(tobj):
    result = Info(tobj.name, tobj.value)
    return result


def info_to_db(obj, tcls):
    result = tcls(obj.name, obj.value)
    return result


def info_table(base):
    class Info(base):
        __tablename__ = 'Info'

        id_ = Column(Integer, Sequence('info_id_seq', metadata=base.metadata), primary_key=True)
        run_id = Column(String, nullable=False)
        name = Column(String, nullable=False)
        value = Column(String, nullable=True)

        __table_args__ = (
                UniqueConstraint('run_id', 'name'),
                )

        def __repr__(self):
            return "<Info(run_id='%s', name='%s', value='%s')>" % (self.run_id, self.name, self.value)

    return Info


Trigger = namedlist("Trigger", "id_ run_id event_id sequence recovery created acted")


def trigger_from_db(tobj):
    result = Trigger(tobj.id_, tobj.run_id, tobj.event_id, tobj.sequence, tobj.recovery, tobj.created, tobj.acted)
    return result


def trigger_to_db(obj, tcls):
    result = tcls(obj.id_, obj.run_id, obj.eveny_id, obj.sequence, obj.recovery, obj.created, obj.acted)
    return result


def trigger_table(base):
    class Trigger(base):
        __tablename__ = 'Trigger'

        id_ = Column(Integer, Sequence('trigger_id_seq', metadata=base.metadata), primary_key=True)
        run_id = Column(String, nullable=False, default='')
        event_id = Column(String, nullable=False)
        sequence = Column(String, nullable=False)
        recovery = Column(Integer, nullable=False, default=0)
        created = Column(DateTime(), default=datetime.utcnow)
        acted = Column(DateTime(), nullable=True)

        def __repr__(self):
            return "<Trigger(id='%s', run_id='%s', event_id='%s', sequence='%s', recovery='%s', created='%s', acted='%s')>" % (
                self.id_, self.run_id, self.event_id, self.sequence, self.recovery, self.created, self.acted)

    return Trigger


Task = namedlist("Task", "id_ run_id step_id sequence host recovery pid status result created updated") 


def task_from_db(tobj):
    result = Task(tobj.id_, tobj.run_id, tobj.step_id, tobj.sequence, tobj.host, tobj.recovery, tobj.pid, tobj.status, tobj.result, tobj.created, tobj.updated)
    return result


def task_to_db(obj, tcls):
    result = tcls(obj.id_, obj.run_id, obj.step_id, obj.sequence, obj.host, obj.recovery, obj.pid, obj.status, obj.result, obj.created, obj.updated)
    return result


def task_table(base):
    class Task(base):
        __tablename__ = 'Task'

        id_ = Column(Integer, Sequence('task_id_seq', metadata=base.metadata), primary_key=True)
        run_id = Column(String, default='', )
        step_id = Column(String, )
        sequence = Column(String, )
        recovery = Column(Integer, nullable=False, )
        host = Column(String, nullable=False, default='')
        pid = Column(Integer, nullable=True)
        status = Column(SQLEnum(TaskStatus), ) 
        result = Column(PickleType() , nullable=True,)
        created = Column(DateTime(), default=datetime.utcnow) 
        updated = Column(DateTime(), nullable=True, ) 

        __table_args__ = (
                UniqueConstraint('run_id', 'step_id', 'sequence', 'recovery'),
                )

        def __repr__(self):
            return "<Task(id='%s', run_id='%s', step_id='%s', sequence='%s', host='%s', recovery='%s', pid='%s', status='%s', created='%s', updated='%s')>" % (
                self.id_, self.run_id, self.step_id, self.sequence, self.host, self.recovery, self.pid, self.status, self.created, self.updated)

    return Task


Delay = namedlist("Delay", "id_ run_id delay_id sequence recovery seconds active activated updated")


def delay_from_db(tobj):
    result = Delay(tobj.id_, tobj.run_id, tobj.delay_id, tobj.sequence, tobj.recovery, tobj.seconds, tobj.active, tobj.activated, tobj.updated)
    return result


def delay_to_db(obj, tcls):
    result = tcls(obj.id_, obj.run_id, obj.delay_id, obj.sequence, obj.recovery, obj.active, obj.activated, obj.updated)
    return result


def delay_table(base):
    class Delay(base):
        __tablename__ = 'Delay'

        id_ = Column(Integer, Sequence('delay_id_seq', metadata=base.metadata), primary_key=True)
        run_id = Column(String, default='')
        delay_id = Column(String,)
        sequence = Column(String,)
        recovery = Column(Integer,)
        seconds = Column(BigInteger, nullable=True)
        active = Column(Boolean, default=False)
        activated = Column(DateTime(), nullable=True,)
        updated = Column(DateTime(), default=datetime.utcnow)

        __table_args__ = (
                UniqueConstraint('run_id', 'delay_id', 'sequence', 'recovery'),
                )

        def __repr__(self):
            return "<Delay(id='%s', run_id='%s', delay_id='%s', sequence='%s', delay='%s', active='%s', activated='%s')>" % (
                self.id_, self.run_id, self.delay_id, self.sequence, self.seconds, self.active, self.activated)

    return Delay


if __name__ == '__main__':
    # engine = create_engine('sqlite:///:memory:', echo=True)
    file = '/var/acrisel/sand/eventor/eventor/eventor/eventor/schema.db'
    dns = 'sqlite:///' + file
    try:
        import os
        os.remove(file)
    except FileNotFoundError:
        pass

    Base = declarative_base()
    engine = create_engine(dns, echo=True)
    Info = info_table(Base)
    Trigger = trigger_table(Base)
    Task = task_table(Base)
    Delay = delay_table(Base)
    Base.metadata.create_all(engine)

    '''
    from sqlalchemy import MetaData
    from sqlalchemy_schemadisplay import create_schema_graph

    graph = create_schema_graph(
      metadata = MetaData(dns),
      show_datatypes=True, # The image would get nasty big if we'd show the datatypes
      show_indexes=True, # ditto for indexes
      #rankdir='LR', # From left to right (instead of top to bottom)
      concentrate=True # Don't try to join the relation lines together
    )

    graph.write_png('schema.png')
    '''

    from sqlalchemy.orm import sessionmaker

    Session = sessionmaker(bind=engine)
    session = Session()

    db_task = Task(step_id='34', status=TaskStatus.ready, recovery='0', sequence='45')
    session.add(db_task)
    session.flush()
    session.commit()

    db_trigger = Trigger(event_id='46', sequence='45')
    session.add(db_trigger)
    session.flush()
    session.commit()

    db_info = Info(run_id='123', name='46', value='45')
    session.add(db_info)
    session.flush()
    session.commit()

    db_delay = Delay(run_id='123', delay_id='S1', recovery=3)
    session.add(db_delay)
    session.flush()
    session.commit()

    db_delay = Delay(delay_id='mydelay', seconds=2419200, sequence=3)
    print('DBAPI - add_delay: %s' % (repr(db_delay), ))
    session.add(db_delay)
    session.flush()
    session.commit()




# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
"""

About
=========
:synopsis:     example use of grapior
:moduleauthor: Arnon Sela
:date:         Oct 18, 2016
:description:  use gradior dependencies and recovery
   
Outputs:
-------------------
N/A

Dependencies:
-------------------
N/A
      
**History:**
-------------------

:Author: Arnon Sela
:Modification:
   - Initial entry
:Date: Oct 18, 2016


API DOC:
===============     
"""

import eventor as evr
import logging
import os
import time
from eventor.examples.run_types import prog

logger = logging.getLogger(__name__)

'''
def prog(progname):
    logger.info("doing what %s is doing" % progname)
    logger.info("EVENTOR_STEP_SEQUENCE: %s" % os.getenv("EVENTOR_STEP_SEQUENCE"))
    return progname
'''


def build_flow(run_mode):
    ev = evr.Eventor(name=os.path.basename(__file__), run_mode=run_mode, logging_level=logging.DEBUG,)

    ev1s = ev.add_event('run_step1')
    ev2s = ev.add_event('run_step2')
    ev3s = ev.add_event('run_step3')

    s1 = ev.add_step('s1', func=prog, kwargs={'progname': 'prog1'}, triggers={evr.StepStatus.success: (ev2s,)}) 
    s2 = ev.add_step('s2', func=prog, kwargs={'progname': 'prog2'}, triggers={evr.StepStatus.success: (ev3s,)})
    s3 = ev.add_step('s3', func=prog, kwargs={'progname': 'prog3'},)

    ev.add_assoc(ev1s, s1, delay=0)
    ev.add_assoc(ev2s, s2, delay=10)
    ev.add_assoc(ev3s, s3, delay=10)

    ev.trigger_event(ev1s, 1)
    return ev


def construct_and_run_in_steps():
    ev = build_flow(run_mode=evr.RunMode.restart)
    ev.run(max_loops=1)
    ev.close()

    for loop in range(4):
        delay = 5 if loop in [1, 2] else 15
        time.sleep(delay)
        ev = build_flow(run_mode=evr.RunMode.continue_)
        result = ev.run(max_loops=1)
        ev.close()
        print('Result: %s' % result)


def construct_and_run():
    ev = build_flow(run_mode=evr.RunMode.restart)
    result = ev.run()
    print('Result: %s' % result)


if __name__ == '__main__':
    import multiprocessing as mp
    mp.freeze_support()
    mp.set_start_method('spawn')
    construct_and_run()





'''
Created on Nov 23, 2016

@author: arnon
'''

from enum import Enum


class EventorError(Exception):
    pass


class AssocType(Enum):
    event = 1
    step = 2


class TaskStatus(Enum):
    ''' Eventor internal task states
    '''
    ready = 1     # triggered to run by events, waiting for resources
    allocate = 2  # resource allocation, waiting for allocation of resources
    fueled = 3   # resource satisfied, waiting to be activated
    active = 4    # running
    success = 5   # finished successfully
    failure = 6   # failed to finish


class StepStatus(Enum):
    ''' Programmer Step interface
    '''
    ready = 1
    allocate = 2
    fueled = 3
    active = 4
    success = 5
    failure = 6
    complete = 7  # finished successfully or with failure


def task_to_step_status(status):
    value = status.value
    result = StepStatus(value)
    return result


def step_to_task_status(status):
    '''
    This assumes that complete status was already converted to failure and success
    '''
    value = status.value
    result = TaskStatus(value)
    return result


class StepReplay(Enum):
    rerun = 1
    skip = 2


class RunMode(Enum):
    restart = 1
    recover = 2
    continue_ = 3
    # replay = 4


class DbMode(Enum):
    write = 1
    append = 2
    read = 3


class LoopControl(Enum):
    stop = 1
    pause = 2
    start = 3
    resume = 4
    kill = 5


class Invoke(object):
    '''
    Invoke is made to support looping super tasks.
    '''
    def __init__(self, target, args=(), kwargs={}):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.is_alive_flag = False

    def start(self,):
        self.is_alive_flag = True
        result = self.target(*self.args, **self.kwargs)
        self.is_alive_flag = False
        return result

    # @threaded
    def join(self, timeout=None):
        return

    # @threaded
    def is_alive(self):
        return self.is_alive_flag



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import eventor as evr
import logging
import os

appname = os.path.basename(__file__)


def prog(progname):
    mlogger = logging.getLogger(os.getenv("EVENTOR_LOGGER_NAME"))
    mlogger.info("doing what %s is doing" % progname)
    mlogger.info("EVENTOR_STEP_SEQUENCE: %s" % os.getenv("EVENTOR_STEP_SEQUENCE"))
    return progname


def build_flow(run_mode, run_id=None):
    ev = evr.Eventor(name=appname,
                     run_mode=run_mode,
                     run_id=run_id,
                     config={'LOGGING':
                             {'logging_level': logging.INFO}})

    ev1s = ev.add_event('run_step1')
    ev2s = ev.add_event('run_step2')
    ev3s = ev.add_event('run_step3')

    kwargs1 = {'progname': 'prog1'}
    kwargs2 = {'progname': 'prog1'}
    kwargs3 = {'progname': 'prog1'}

    s1 = ev.add_step('s1', func=prog, kwargs=kwargs1, triggers={evr.STEP_SUCCESS: (ev2s,)})
    s2 = ev.add_step('s2', func=prog, kwargs=kwargs2, triggers={evr.STEP_SUCCESS: (ev3s,)})
    s3 = ev.add_step('s3', func=prog, kwargs=kwargs3,)

    ev.add_assoc(ev1s, s1, delay=0)
    ev.add_assoc(ev2s, s2, delay=0)
    ev.add_assoc(ev3s, s3, delay=0)

    ev.trigger_event(ev1s, 1)
    return ev


if __name__ == '__main__':
    ev = build_flow(run_mode=evr.RUN_RESTART)
    ev.run()



'''
Created on Oct 31, 2016

@author: arnon
'''

from .eventor_types import EventorError
import logging
from _ast import arg
import collections

mlogger = logging.getLogger(__name__)


def expr_to_str(*args):
    items = list()
    for arg in args:
        if type(arg) == str:
            result = arg
        elif isinstance(arg, Event):
            result = arg.id_
        elif isinstance(arg, collections.Iterable):
            result = "(" + expr_to_str(*arg) + ")"
        else:
            raise EventorError("unknown variable in logical operation: %s" % repr(arg))
        items.append(result)

    expr = "(" + ") and (".join(items) + ")"
    return expr


def or_(*args):
    items = list()
    for arg in args:
        result = expr_to_str(arg)
        items.append(result)

    expr = " or ".join(items)
    return expr


class Event(object):
    """Event object used to reuse in user program.

        Event objects are created by Eventor.add_event programming interface or by the mechanism itself

        Attributes:
            N/A

        Methods:
            db_write: write event to db file

    """

    def __init__(self, name, expr=None, logger=None):
        global mlogger

        if logger:
            mlogger = logger
        self.name = name
        self.expr = expr
        if expr:
            self.expr = expr_to_str(expr)
        self.id_ = name  # get_event_id()

    def __repr__(self):
        expr = ''
        if self.expr:
            expr = ", %s" % self.expr
        return "Event(%s%s)" % (repr(self.id_), expr)

    def __str__(self):
        return repr(self)

    # def db_write(self, db):
    #     db.add_event(event_id=self.id_, name=self.name,)

    def trigger_(self, db, sequence):
        mlogger.debug('[ Event {}/{} ] Adding as trigger to DB'.format(self.name, sequence,))
        db.add_trigger(event_id=self.id_, sequence=sequence,)

    def trigger_if_not_exists(self, db, sequence, recovery):
        mlogger.debug('[ Event {}/{} ] Adding, if not already exists, as trigger to DB'
                      .format(self.name, sequence,))
        try:
            added = db.add_trigger_if_not_exists(event_id=self.id_, sequence=sequence,
                                                 recovery=recovery)
        except Exception:
            # mlogger.exception(e)
            raise
        return added

    def add_expr(self, expr):
        ''' override with new expr
        '''
        if expr:
            self.expr = expr_to_str(expr)


if __name__ == '__main__':
    e1 = Event('E1')
    e2 = Event('E2')
    e3 = Event('E3')
    e4 = Event('E4')

    expr = expr_to_str(or_(e1, e2), or_(e3, e4))
    print(expr, eval(expr, globals(),
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True}))

    expr = expr_to_str(e3, e4)

    print(expr, eval(expr, globals(),
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True}))

    expr = expr_to_str((e1, e2))
    print(expr, eval(expr, globals(),
                     {'E0': True, 'E1': True, 'E2': False, 'E3': False, 'E4': True}))
